# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .providers import configs_factory
from . import ConfigKeyConst


_settings = configs_factory()


def format_event_type(type):
    return "{}{}".format(_settings.get(ConfigKeyConst.TYPE_TOPICS_PREFIX, ""), type)


_m = format_event_type


SUBJECT_PROPERTY_CHANGED = _m("subject-property-changed")
SUBJECT_PROPERTY_DELETED = _m("subject-property-deleted")
